require 'display_triangle.rb'
require 'pattern_canvas.rb'

# -------------------------------------------------------
# --- demo display class
# -------------------------------------------------------

class TriangleMargeDemo < MYGLWindowViewTriangle

  def initialize(width, height, title)
    @canvas = PatternStorage.new
    @tri1 = PTriangle.new
    @tri2 = PTriangle.new
    @tri1pos = PPoint.new(0,0,0)
    @tri2pos = PPoint.new(0,0,0)
    @tick_count = 0
    @outline = false
    #
    super
  end

  def mouse_down(e)
    super
    if e.Button.to_s == 'Right' then
      @outline = !@outline
    end
  end

  def tick
    move_triangle
    @canvas.clear
    @canvas.add(@tri1)
    @canvas.add(@tri2)
    @canvas.drip_outline
    #
    clear_object()
    if @outline == true then
      @canvas.outline.each do |line|
        add_line(line)
      end
    else
      @canvas.object.each do |line|
        add_object(line)
      end
    end
    #
    @tick_count += 1
    refresh
  end

  def move_triangle
    p1 = PPoint.new(-1.0, -0.8, 0)
    p2 = PPoint.new( 0.2,  0.7, 0)
    p3 = PPoint.new( 0.9, -0.9, 0)
    p4 = PPoint.new( 0.8,  0.5, 0)
    p5 = PPoint.new( 0.2, -0.7, 0)
    p6 = PPoint.new(-0.7,  0.4, 0)
    @tri1.set_point(0, p1)
    @tri1.set_point(1, p2)
    @tri1.set_point(2, p3)
    @tri1.normalize
    @tri2.set_point(0, p4)
    @tri2.set_point(1, p5)
    @tri2.set_point(2, p6)
    @tri2.normalize
    #
    r1 = 1.0 * @tick_count
    r2 = -1.7 * @tick_count
    @tri1.get_point(0).rotate_z(r1)
    @tri1.get_point(1).rotate_z(r1)
    @tri1.get_point(2).rotate_z(r1)
    @tri2.get_point(0).rotate_z(r2)
    @tri2.get_point(1).rotate_z(r2)
    @tri2.get_point(2).rotate_z(r2)
    #
    m1 = 0.003 * @tick_count
    m2 = -0.00183 * @tick_count
    m3 = -0.0009 * @tick_count
    m4 = 0.0021 * @tick_count
    mv1 = PPoint.new( 1.0*Math.cos(m1), 0.5*Math.sin(m3), 0)
    mv2 = PPoint.new( 1.0*Math.cos(m2), 0.5*Math.sin(m4), 0)
    @tri1.get_point(0).add(mv1)
    @tri1.get_point(1).add(mv1)
    @tri1.get_point(2).add(mv1)
    @tri2.get_point(0).add(mv2)
    @tri2.get_point(1).add(mv2)
    @tri2.get_point(2).add(mv2)
  end

end



# -------------------------------------------------------
# --- boot & main
# -------------------------------------------------------

demo = TriangleMargeDemo.new(800, 600, "Triangle marge demo")
demo.running