require "draw_primitive.rb"

# --------------------------------------------------------
#  triangle marge work
# --------------------------------------------------------

class TriangleCollision
  def initialize(t1)
    @tri1 = t1
    @div1 = []
    @div2 = []
    @epoint1 = []
    @epoint2 = []
    @epoint3 = []
  end

  attr_accessor :div1
  attr_accessor :div2

  def marge(tri2)
    divide_crossline(tri2)
  end

  def build_div1()
    @div1.clear
    subdivision
  end

  # ------------------------------
  # --- tool method
  def include?(tri1, tri2)
    p1 = tri2.get_point(0)
    if (tri1.include?(p1) == false) && (onedge?(tri1, p1) == false) then
      return false
    end
    p2 = tri2.get_point(1)
    if (tri1.include?(p2) == false) && (onedge?(tri1, p2) == false) then
      return false
    end
    p3 = tri2.get_point(2)
    if (tri1.include?(p3) == false) && (onedge?(tri1, p3) == false) then
      return false
    end
    return true    
  end

  def onedge?(tri, point)
    line = PLine.new
    line.set_point(0, tri.get_point(0))
    line.set_point(1, tri.get_point(1))
    if online?(line, point) then
  return true
    end
    line.set_point(0, tri.get_point(1))
    line.set_point(1, tri.get_point(2))
    if online?(line, point) then
      return true
    end
    line.set_point(0, tri.get_point(2))
    line.set_point(1, tri.get_point(0))
    if online?(line, point) then
      return true
    end
    return false
  end

  def crossbase(line1, line2)
    result = Hash.new
    l1p1 = line1.get_point(0)
    l1p2 = line1.get_point(1)
    l2p1 = line2.get_point(0)
    l2p2 = line2.get_point(1)
    det = (l2p2.x - l2p1.x)*(l1p2.y - l1p1.y) - (l1p2.x - l1p1.x)*(l2p2.y - l2p1.y)
    if det == 0 then
      return nil
    end
    dx = l2p1.x - l1p1.x
    dy = l2p1.y - l1p1.y
    t1 = ((l2p2.x - l2p1.x)*dy - (l2p2.y - l2p1.y)*dx) / det
    t2 = ((l1p2.x - l1p1.x)*dy - (l1p2.y - l1p1.y)*dx) / det
    t1 = 0 if t1 < 0.00000000001 && t1 > 0
    t2 = 0 if t2 < 0.00000000001 && t1 > 0
    t1 = 1.0 if t1 > 0.999999999 && t1 < 1.0
    t2 = 1.0 if t2 > 0.999999999 && t1 < 1.0
    result['t1'] = t1
    result['t2'] = t2
    return result
  end

  def crosspoint(line1, line2)
    t = crossbase(line1, line2)
    l1p1 = line1.get_point(0)
    l1p2 = line1.get_point(1)
    l2p1 = line2.get_point(0)
    l2p2 = line2.get_point(1)
    crosspoint = PPoint.new
    crosspoint.x = l1p1.x + (l1p2.x - l1p1.x) * t['t1']
    crosspoint.y = l1p1.y + (l1p2.y - l1p1.y) * t['t1']
    crosspoint.z = 0
    return crosspoint
  end

  def cross?(line1, line2)
    t = crossbase(line1, line2)
    if t == nil then
      return false
    end
    return ((t['t1'] > 0) && (t['t1'] < 1.0) && (t['t2'] > 0) && (t['t2'] < 1.0))
  end

  def cross_with_top?(line1, line2)
    t = crossbase(line1, line2)
    if t == nil then
      return false
    end
    return ((t['t1'] >= 0) && (t['t1'] <= 1.0) && (t['t2'] >= 0) && (t['t2'] <= 1.0))
  end

  def online?(line, point)
    l1p1 = line.get_point(0)
    l1p2 = line.get_point(1)
    if l1p1.x == l1p2.x then
      if (point.x == l1p1.x) then
        t = (point.y - l1p1.y) / (l1p2.y - l1p1.y)
        if (t >= 0) && (t <= 1) then
          return true
        else
          return false
        end
      else
        return false
      end
    end
    #
    t1 = (point.x - l1p1.x) / (l1p2.x - l1p1.x)
    t2 = (point.y - l1p1.y) / (l1p2.y - l1p1.y)
    if (t1 < 0) || (t1 > 1) || (t2 < 0) || (t2 > 1) then
      return false
    end
    if (t1 - t2).abs > 0.000001 then
      return false
    end
    return true
  end

  def collision?(tri1, tri2)
    line1 = PLine.new
    line2 = PLine.new
    line3 = PLine.new
    line4 = PLine.new
    line1.set_point(0, tri1.get_point(0))
    line1.set_point(1, tri1.get_point(1))
    line2.set_point(0, tri2.get_point(0))
    line2.set_point(1, tri2.get_point(1))
    if cross?(line1, line2) then
      return true
    end
    line3.set_point(0, tri2.get_point(1))
    line3.set_point(1, tri2.get_point(2))
    if cross?(line1, line3) then
      return true
    end
    line4.set_point(0, tri2.get_point(2))
    line4.set_point(1, tri2.get_point(0))
    if cross?(line1, line4) then
      return true
    end
    line1.set_point(0, tri1.get_point(1))
    line1.set_point(1, tri1.get_point(2))
    if cross?(line1, line2) then
      return true
    end
    if cross?(line1, line3) then
      return true
    end
    if cross?(line1, line4) then
      return true
    end
    line1.set_point(0, tri1.get_point(2))
    line1.set_point(1, tri1.get_point(0))
    if cross?(line1, line2) then
      return true
    end
    if cross?(line1, line3) then
      return true
    end
    if cross?(line1, line4) then
      return true
    end
    return false
  end

  # -----------------------------------------
  # divide work
  def divide_crossline(tri2)
    div1 = []
    div2 = []
    # -- include
    if include?(@tri1, tri2) then
      @div1 << @tri1
      return true
    end 
    if include?(tri2, @tri1) then
      @div2 << tri2
      return false
    end 
    # -- not collision
    if collision?(@tri1, tri2) == false then
      @div1 << @tri1
      @div2 << tri2
      return false
    end

    # --- divide work
    # -- is tri1 point in tri2?
    if subdivision_nest(tri2) then
      @div1.clear()
      subdivision
      return true
    end

    # -- is tri2 point on tri1 edge?
    if cut(tri2) then
      @div1.clear()
      subdivision
      return true
    end

    # -- normal marge
    if clip(tri2) == false then
      @div2 << tri2
    end
    @div1.clear()
    subdivision

    # -- next
    @div1 += div1
    @div2 += div2
    return true
  end

  def subdivision_nest(tri2)
    includepoint = false
    for i in 0...3 do
      t1point = @tri1.get_point(i)
      if tri2.include?(t1point) &&
         tri2.get_point(0).equal?(t1point)==false &&
         tri2.get_point(1).equal?(t1point)==false &&
         tri2.get_point(2).equal?(t1point)==false then
        includepoint = true
        # - do nesting
        for j in 0...3 do
          tri = PTriangle.new
          tri.set_point(0, t1point)
          tri.set_point(1, tri2.get_point(j%3))
          tri.set_point(2, tri2.get_point((j+1)%3))
          if include?(@tri1, tri) == false then
            divide_crossline(tri)
          end
        end
        break
      end
    end
    return includepoint
  end

  def cut(tri2)
    onpoint = false
    #
    for i in 0...3 do
      t2point = tri2.get_point(i)
      for j in 0...3 do
        line1 = PLine.new
        line1.set_point(0, @tri1.get_point(j%3))
        line1.set_point(1, @tri1.get_point((j+1)%3))
        if online?(line1, t2point)  then
          line2 = PLine.new
          line2.set_point(0, tri2.get_point((i+1)%3))
          line2.set_point(1, tri2.get_point((i+2)%3))
          if cross?(line1, line2) then
            p = crosspoint(line1, line2)
            tri = PTriangle.new
            tri.set_point(0, p)
            tri.set_point(1, t2point)
            tri.set_point(2, tri2.get_point((i+2)%3))
            tri.normalize
            divide_crossline(tri)
            tri = PTriangle.new
            tri.set_point(0, tri2.get_point((i+1)%3))
            tri.set_point(1, t2point)
            tri.set_point(2, p)
            tri.normalize
            divide_crossline(tri)
            #
            case j
            when 0
              @epoint1 << t2point
              @epoint1 << p
            when 1
              @epoint2 << t2point
              @epoint2 << p
            when 2
              @epoint3 << t2point
              @epoint3 << p
            end
            onpoint = true
          end
        end
      end
    end
    return onpoint
  end

  def clip(tri2)
    clipping_flag = false
    for i in 0...3 do
      for j in 0...3 do
        cutter = PLine.new
        cutter.set_point(0, @tri1.get_point(j))
        cutter.set_point(1, @tri1.get_point((j+1)%3))
        e1 = PLine.new
        e1.set_point(0, tri2.get_point(i))
        e1.set_point(1, tri2.get_point((i+1)%3))
        e2 = PLine.new
        e2.set_point(0, tri2.get_point((i+1)%3))
        e2.set_point(1, tri2.get_point((i+2)%3))
        p1 = tri2.get_point((i+1)%3)
        #
        next if (cross?(cutter, e1) == false) || (cross?(cutter, e2) == false)
        #
        p2 = crosspoint(cutter, e2)
        p3 = crosspoint(cutter, e1)
        p4 = tri2.get_point(i)
        p5 = tri2.get_point((i+2)%3)
        ccheck1 = PLine.new
        ccheck1.set_point(0, p1)
        ccheck1.set_point(1, p2)
        ccheck2 = PLine.new
        ccheck2.set_point(0, @tri1.get_point(0))
        ccheck2.set_point(1, @tri1.get_point(1))
        ccheck3 = PLine.new
        ccheck3.set_point(0, @tri1.get_point(1))
        ccheck3.set_point(1, @tri1.get_point(2))
        ccheck4 = PLine.new
        ccheck4.set_point(0, @tri1.get_point(2))
        ccheck4.set_point(1, @tri1.get_point(0))
        if @tri1.include?(p1)||
           cross?(ccheck1, ccheck2) ||
           cross?(ccheck1, ccheck3) ||
           cross?(ccheck1, ccheck4) then
          cutout1 = PTriangle.new
          cutout1.set_point(0, p2)
          cutout1.set_point(1, p3)
          cutout1.set_point(2, p4)
          cutout1.normalize()
          if collision?(@tri1, cutout1) == false then
            @div2 << cutout1
            clipping_flag = true
          end
          cutout2 = PTriangle.new
          cutout2.set_point(0, p4)
          cutout2.set_point(1, p5)
          cutout2.set_point(2, p2)
          cutout2.normalize()
          if collision?(@tri1, cutout2) == false then
            @div2 << cutout2
            clipping_flag = true
          end
        else
          cutout = PTriangle.new
          cutout.set_point(0, p1)
          cutout.set_point(1, p2)
          cutout.set_point(2, p3)
          @div2 << cutout
          clipping_flag = true
        end
        #
        case j
        when 0
          @epoint1 << p2
          @epoint1 << p3
        when 1
          @epoint2 << p2
          @epoint2 << p3
        when 2
          @epoint3 << p2
          @epoint3 << p3
        end
      end
    end
    return clipping_flag
  end

  def subdivision
    @epoint1.sort! do |a,b|
      [a.x, a.y] <=> [b.x, b.y]
    end
    @epoint2.sort! do |a,b|
      [a.x, a.y] <=> [b.x, b.y]
    end
    @epoint3.sort! do |a,b|
      [a.x, a.y] <=> [b.x, b.y]
    end
    #
    side1 = PTriangle.new
    side2 = PTriangle.new
    side1.set_point(0, @tri1.get_point(0))
    side1.set_point(1, @tri1.get_point(1))
    side1.set_point(2, @tri1.get_point(2))
    side2.set_point(0, @tri1.get_point(0))
    side2.set_point(1, @tri1.get_point(1))
    side2.set_point(2, @tri1.get_point(2))
    #
    if @epoint1.size > 0 then
      first = true
      dirc = subdiv_direction(@tri1, @epoint1)
      top = dirc['top']
      ppos = dirc['start']
      @epoint1.each do |ep|
        next if @tri1.get_point(0).equal?(ep)
        next if @tri1.get_point(1).equal?(ep)
        next if @tri1.get_point(2).equal?(ep)
        if first == true then
          side2.set_point(0, top)
          side2.set_point(1, ppos)
          side2.set_point(2, ep)
        else
          at = PTriangle.new
          at.set_point(0, top)
          at.set_point(1, ppos)
          at.set_point(2, ep)
          @div1 << at
        end
        ppos = ep
        first = false
      end
      side1.set_point(0, top)
      side1.set_point(1, ppos)
      side1.set_point(2, dirc['end'])
    end
    # 
    if @epoint2.size > 0 then
      first = true
      dirc = subdiv_direction(side1, @epoint2)
      if dirc.key?('top') == false then
        t = side1
        side1 = side2
        side2 = t
        dirc = subdiv_direction(side1, @epoint2)
      end
      top = dirc['top']
      ppos = dirc['start']
      @epoint2.each do |ep|
        next if @tri1.get_point(0).equal?(ep)
        next if @tri1.get_point(1).equal?(ep)
        next if @tri1.get_point(2).equal?(ep)
        if (first == true) and (@epoint1.size == 0) then
          side2.set_point(0, top)
          side2.set_point(1, ppos)
          side2.set_point(2, ep)
        else
          at = PTriangle.new
          at.set_point(0, top)
          at.set_point(1, ppos)
          at.set_point(2, ep)
          at.normalize
          @div1 << at
        end
        ppos = ep
        first = false
      end
      at = PTriangle.new
      at.set_point(0, top)
      at.set_point(1, ppos)
      at.set_point(2, dirc['end'])
      @div1 << at
    else
      @div1 << side1 if @epoint1.size > 0
    end
    #
    if @epoint3.size > 0 then
      first = true
      dirc = subdiv_direction(side2, @epoint3)
      if dirc.key?('top') == false then
        t = side1
        side1 = side2
        side2 = t
        dirc = subdiv_direction(side2, @epoint2)
      end
      top = dirc['top']
      ppos = dirc['start']
      @epoint3.each do |ep|
        next if @tri1.get_point(0).equal?(ep)
        next if @tri1.get_point(1).equal?(ep)
        next if @tri1.get_point(2).equal?(ep)
        at = PTriangle.new
        at.set_point(0, top)
        at.set_point(1, ppos)
        at.set_point(2, ep)
        @div1 << at
        ppos = ep
      end
      at = PTriangle.new
      at.set_point(0, top)
      at.set_point(1, ppos)
      at.set_point(2, dirc['end'])
      @div1 << at
    else
      @div1 << side2
    end
    #
  end

  def subdiv_direction(tri, divpoints)
    result = Hash.new
    point = nil
    divpoints.each do |dp|
      r = false
      for i in 0...3
        r = true if tri.get_point(i).equal?(dp)
      end
      next if r == true
      point = dp
      break
    end
    if point == nil then
      #failsafe
      result["top"] = tri.get_point(0)
      result["start"] = tri.get_point(1)
      result["end"] = tri.get_point(2)
      return result
    end
    #
    for i in 0...3
      line = PLine.new
      line.set_point(0, tri.get_point(i))
      line.set_point(1, tri.get_point((i+1)%3))
      if online?(line, point) then
        result['top'] = tri.get_point((i+2)%3)
        p1 = tri.get_point(i)
        p2 = tri.get_point((i+1)%3)
        if p1.x == p2.x then
          if p1.y <= p2.y then
            result['start'] = p1
            result['end'] = p2
          else
            result['start'] = p2
            result['end'] = p1
          end
        else
          if p1.x <= p2.x then
            result['start'] = p1
            result['end'] = p2
          else
            result['start'] = p2
            result['end'] = p1
          end
        end
        #
        break
      end
    end
    return result
  end

end


# --------------------------------------------------------
#  pattern stored class
# --------------------------------------------------------

class PatternStorage

  def initialize
    @object = []
    @outline = []
  end

  attr_reader :object
  attr_reader :outline

  def object_size?
    @object.size
  end

  def clear
    @object.clear
    @outline.clear
  end

  def add( triangle )
    triangle.normalize()
    if @object.size == 0 then
      @object << triangle
    else
      marge( triangle )
    end
  end

  def drip_outline()
    @outline = []
    make_outline()
  end

  # ------------------------------
  # --- object marge work
  def marge( addtri )
    @delete_index = []
    count = @object.size
    # -- add
    margework = TriangleCollision.new(addtri)
    for i in 0...count
      if margework.marge(@object[i]) == true then
        @delete_index << i
        @object += margework.div2
      end
    end
    margework.build_div1()
    @object += margework.div1
    # -- delete
    @delete_index.each do |index|
      @object.delete_at(index)
    end
  end

  # ------------------------------
  # --- extruct object outline
  class PLineCount < PLine
    def initialize
      super
      @count = 0
    end

    attr_accessor :count
  end

  def make_outline()
    @outline = []
    linelist = []
    @object.each do |tri|
      for i in 0...3
        p1 = tri.get_point(i)
        p2 = tri.get_point((i+1)%3)
        line = PLineCount.new
        line.set_point(0, p1)
        line.set_point(1, p2)
        check = false
        linelist.each do |ol|
          if ol.equal?(line) then
            ol.count += 1
            check = true
            break
          end
        end
        linelist << line if check==false
      end
    end
    # - pickup outline
    linelist.each do |ol|
      if ol.count == 0 then
        line = PLine.new
        line.set_point(0, ol.get_point(0))
        line.set_point(1, ol.get_point(1))
        @outline << line
      end
    end
  end

end

# --------------------------------------------------------
#  triangle marge work
# --------------------------------------------------------

class PatternCanvas

  def initialize
    @pattern = PatternStorage.new
  end


end